/****************************************************************************
 *                                                                          *
 *                  Concours 2006 - GAMES CREATORS NETWORK                  *
 *                                                                          *
 ****************************************************************************

                     PRODUCT          : GRAVITY GAME
 
                     FILE             : SoundManager.cpp

                     VERSION          : 1
 
 ****************************************************************************
 *                                                                          *
 *  DESCRIPTION :                                                           *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          * 
 *                                                                          *  
 ****************************************************************************
 *                                                                          *
 * HISTORY :                                                                * 
 * ---------                                                                *
 *                                                                          *
 *  Date        | Author  | Comment                                         *
 *  ---------------------------------------------------------------------   *
 *  24/07/2006  | Bahamut | Creation                                        *
 *  ---------------------------------------------------------------------   *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

#include "SoundManager.h"

/****************************************************************************
 *                                                                          *
 *    Function     : CSoundManager                                          * 
 *                                                                          *
 *    Description  : Class construtor                                       *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
CSoundManager::CSoundManager(void)
: m_bInited(false)
, m_iLoopingMusic(false)
, m_iAudioFrequency(65535)
, m_iAudioFormat(16)
, m_iAudioChannelsNumber(64)
, m_pMusicStream(0)
, m_iMusicChannel(0)
, m_iCurrentVolume(70)
{
	m_eError = SND_NO_ERROR;
}
 

/****************************************************************************
 *                                                                          *
 *    Function     : ~CSoundManager                                         * 
 *                                                                          *
 *    Description  : Class construtor                                       *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
CSoundManager::~CSoundManager(void)
{
	if (m_bInited)
		Close();
}
 
 
/****************************************************************************
 *                                                                          *
 *    Function     : Init                                                   * 
 *                                                                          *
 *    Description  : Initializes FMOD Sound library                         *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value : Returns true if succeed, else returns false            *
 *                                                                          *
 ****************************************************************************/
bool CSoundManager::Init(void)
{
	bool result = true;
    if (FSOUND_GetVersion() < FMOD_VERSION)
    {
		// Error: bad version.
		fprintf(stderr, "error: FMOD has bad version. The current version is :%f\n", FMOD_VERSION );
		
		m_eError = SND_ERROR_WRONG_VERSION;
        result = false;
    }
    
	//FSOUND_SetOutput(FSOUND_OUTPUT_DSOUND); // sometimes, it does may don't work on Windows XP (Direct X) bug ?
	FSOUND_SetOutput(FSOUND_OUTPUT_WINMM);
	FSOUND_SetDriver(0);

    if (!FSOUND_Init(m_iAudioFrequency, m_iAudioChannelsNumber, 0))
    {
		// Error: Can't initializes Sound Library. Fatal error.
		fprintf(stderr, "error: FMOD can't be initialized.\n" );
		m_eError = SND_ERROR_INIT;
        result = false;
    }

	if (result)
	{
		// Note: Sound Library ready.
		m_bInited = true;
	}
	return result;
}
 
 
/****************************************************************************
 *                                                                          *
 *    Function     : Close                                                  * 
 *                                                                          *
 *    Description  : DeInitializes FMOD Sound library                       *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CSoundManager::Close(void)
{
	stopSamples();
	UnloadSamples();
	UnloadMusics();

	FSOUND_Close();
	m_bInited = false;
	// Sound library terminated with succeed.
}


/****************************************************************************
 *                                                                          *
 *    Function     : LoadSample                                             * 
 *                                                                          *
 *    Description  : Loads a sample in memory                               *
 *                                                                          *
 *    Interface    : I filename = physical file name                        *
 *                   I name = sound sample alias                            *
 *                                                                          *
 *    Return Value : Returns 0 if succeed, else returns -1                  *
 *                                                                          *
 ****************************************************************************/
int CSoundManager::LoadSample(const char * filename, const char * name)
{
	if ( !m_tSoundTable.count(filename) )
	{
		FSOUND_SAMPLE * sound = FSOUND_Sample_Load(FSOUND_FREE,filename,0, 0,0);
		if (sound)
		{
			m_tSoundTable[name] = sound;
			return 0;
		}
	}
	// Error: Can't load Sample.
	fprintf(stderr, "error: Sound %s (%s) can't be loaded.\n", name,filename );
	return -1;
}
 
 
/****************************************************************************
 *                                                                          *
 *    Function     : LoadMusic                                              * 
 *                                                                          *
 *    Description  : Loads music in memory                                  *
 *                                                                          *
 *    Interface    : I filename = physical file name                        *
 *                   I name = music sample alias                            *
 *                                                                          *
 *    Return Value : Returns 0 if succeed, else returns -1                  *
 *                                                                          *
 ****************************************************************************/
int CSoundManager::LoadMusic(const char * filename, const char * name)
{
	if ( !m_tMusicTable.count(name) )
	{
		unsigned int flags = 0; //FSOUND_HW2D | FSOUND_STEREO | FSOUND_16BITS | FSOUND_MPEGACCURATE;

		FSOUND_STREAM * music = FSOUND_Stream_Open(filename,flags,0,0);
		if (music)
		{
			m_tMusicTable[name] = music;
			return 0;
		}
	}
	// Error: Can't load music
	fprintf(stderr, "error: Music %s (%s) can't be loaded.\n", name,filename );
	return -1;
}
 
 
/****************************************************************************
 *                                                                          *
 *    Function     : PlaySample                                             * 
 *                                                                          *
 *    Description  : Plays sample                                           *
 *                                                                          *
 *    Interface    : I sound = sample alias                                 *
 *                   I loop = loop mode                                     *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CSoundManager::PlaySample(std::string sound, bool loop)
{
	if ( m_tSoundTable.count(sound) )
	{
		int channel = FSOUND_PlaySound(FSOUND_FREE,m_tSoundTable[sound]);

		if (loop)
			FSOUND_SetLoopMode(channel,FSOUND_LOOP_NORMAL);
	}
	else
	{
		// Error: Can't play sample. Sample not loaded.
		fprintf(stderr, "error: Sound %s can't be played.\n", sound.c_str() );
	}
}
 

/****************************************************************************
 *                                                                          *
 *    Function     : UnloadSample                                           * 
 *                                                                          *
 *    Description  : Unloads sample                                         *
 *                                                                          *
 *    Interface    : I sound = sample alias to unload                       *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CSoundManager::UnloadSample(std::string sound)
{
	if ( m_tSoundTable.count(sound) )
	{
		FSOUND_Sample_Free(m_tSoundTable[sound]);
		m_tSoundTable.erase( sound ) ;
	}
	else
	{
		// Error: Can't unload sample. Sample not loaded.
		fprintf(stderr, "error: Sound %s can't be unloaded.\n", sound.c_str() );
	}
}


/****************************************************************************
 *                                                                          *
 *    Function     : UnloadMusic                                            * 
 *                                                                          *
 *    Description  : Unloads music                                          *
 *                                                                          *
 *    Interface    : I music = music alias to unload                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CSoundManager::UnloadMusic(std::string music)
{
	if ( m_tMusicTable.count(music) )
	{
		FSOUND_Stream_Close(m_tMusicTable[music]);
		m_tMusicTable.erase( music ) ;
	}
	else
	{
		// Error: Can't unload music. Music not loaded.
		fprintf(stderr, "error: Music %s can't be loaded.\n", music.c_str() );
	}
}

/****************************************************************************
 *                                                                          *
 *    Function     : UnloadMusics                                           * 
 *                                                                          *
 *    Description  : Unloads alls musics                                    *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CSoundManager::UnloadMusics()
{
	std::map<std::string, FSOUND_STREAM *>::const_iterator itr;
	for(itr = m_tMusicTable.begin(); itr != m_tMusicTable.end(); ++itr)
	{
		FSOUND_Stream_Close((*itr).second);
	}
	m_tMusicTable.clear();
}



/****************************************************************************
 *                                                                          *
 *    Function     : UnloadSamples                                          * 
 *                                                                          *
 *    Description  : Unloads alls samples                                   *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CSoundManager::UnloadSamples()
{
	std::map<std::string, FSOUND_SAMPLE *>::const_iterator itr;
	for(itr = m_tSoundTable.begin(); itr != m_tSoundTable.end(); ++itr)
	{
		FSOUND_Sample_Free((*itr).second);
	}
	m_tSoundTable.clear(); 
}


/****************************************************************************
 *                                                                          *
 *    Function     : stopSamples                                            * 
 *                                                                          *
 *    Description  : Stops alls samples                                     *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CSoundManager::stopSamples(void)
{
	FSOUND_StopSound(FSOUND_ALL);
}

 
 
/****************************************************************************
 *                                                                          *
 *    Function     : PlayMusic                                              * 
 *                                                                          *
 *    Description  : Plays a music                                          *
 *                                                                          *
 *    Interface    : I sound = music alias                                  *
 *                   I loop = loop mode                                     *
 *                                                                          *
 *    Return Value : Returns 0 if succeed, else returns -1                  *
 *                                                                          *
 ****************************************************************************/
void CSoundManager::PlayMusic(std::string music, bool loop)
{
	if ( m_tMusicTable.count(music) )
	{
		if (m_pMusicStream)
			FSOUND_Stream_Stop(m_pMusicStream);

		m_pMusicStream = m_tMusicTable[music];

		if (loop)
			FSOUND_Stream_SetLoopCount(m_pMusicStream,-1);

		m_iMusicChannel = FSOUND_Stream_Play(FSOUND_FREE, m_pMusicStream);
	}
	else
	{
		// Error: Can't play music. Music not loaded.
		fprintf(stderr, "error: music %s can't be played.\n", music.c_str() );
	}
}
 

/****************************************************************************
 *                                                                          *
 *    Function     : PauseMusic                                             * 
 *                                                                          *
 *    Description  : Pauses current music stream                            *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CSoundManager::PauseMusic()
{
	if (m_pMusicStream)
	{
		FSOUND_SetPaused(m_iMusicChannel, true);
	}
}
 
 
/****************************************************************************
 *                                                                          *
 *    Function     : ResumeMusic                                            * 
 *                                                                          *
 *    Description  : Resumes current music stream                           *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CSoundManager::ResumeMusic()
{
	if (m_pMusicStream)
	{
		FSOUND_SetPaused(m_iMusicChannel, false);
	}
}
 
 
/****************************************************************************
 *                                                                          *
 *    Function     : StopMusic                                              * 
 *                                                                          *
 *    Description  : Stops music                                            *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CSoundManager::StopMusic()
{
	if (m_pMusicStream)
	{
		FSOUND_Stream_Stop(m_pMusicStream);
		m_pMusicStream = 0;
	}
}
 

/****************************************************************************
 *                                                                          *
 *    Function     : SetVolume                                              * 
 *                                                                          *
 *    Description  : Sets volume linearly                                   *
 *                                                                          *
 *    Interface    : I volume = The volume to set. Valid ranges are         *
 *                              from 0 (silent) to 255 (full volume)        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CSoundManager::SetVolume(int volume)
{
	if (volume<0)
		volume =0;
	if (volume>255)
		volume = 255;

	m_iCurrentVolume = volume;
    FSOUND_SetVolume(FSOUND_ALL,m_iCurrentVolume);
}
 

/****************************************************************************
 *                                                                          *
 *    Function     : readFile                                               * 
 *                                                                          *
 *    Description  : Loads a textures list in one file(XML Formalism)       *
 *                                                                          *
 *    Interface    : I fileName = physical file                             *
 *                                                                          *
 *    Return Value : Returns true if succeed, else returns false            *
 *                                                                          *
 ****************************************************************************/
bool CSoundManager::readFile(std::string fileName)
{
	TiXmlDocument doc( fileName );
	if ( !doc.LoadFile() )
	{
		// Error: Can't read file.
		fprintf(stderr, "error: Sound file %s can't be parsed.\n", fileName.c_str() );
		return false;
	}

	TiXmlHandle hdl(&doc);
	TiXmlElement *XMLelement = hdl.FirstChild( "Sounds").FirstChildElement().Element();
	
	while (XMLelement)
	{
		std::string name = XMLelement->Attribute("name");
		std::string file = XMLelement->Attribute("file");

		LoadSample(file.c_str(),name.c_str());

		XMLelement = XMLelement->NextSiblingElement();
	}

	XMLelement = hdl.FirstChild( "Musics").FirstChildElement().Element();
	while (XMLelement)
	{
		std::string name = XMLelement->Attribute("name");
		std::string file = XMLelement->Attribute("file");

		LoadMusic(file.c_str(),name.c_str());

		XMLelement = XMLelement->NextSiblingElement();
	}

	return true;
}


/****************************************************************************
 *                                                                          *
 *    Function     : GetError                                               * 
 *                                                                          *
 *    Description  : Returns the current error                              *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value : Returns error state                                    *
 *                                                                          *
 ****************************************************************************/
int CSoundManager::GetError()
{
	return m_eError;
}
