/****************************************************************************
 *                                                                          *
 *                  Concours 2006 - GAMES CREATORS NETWORK                  *
 *                                                                          *
 ****************************************************************************

                     PRODUCT          : GRAVITY GAME
 
                     FILE             : TextureManager.cpp

                     VERSION          : 1
 
 ****************************************************************************
 *                                                                          *
 *  DESCRIPTION :                                                           *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          * 
 *                                                                          *  
 ****************************************************************************
 *                                                                          *
 * HISTORY :                                                                * 
 * ---------                                                                *
 *                                                                          *
 *  Date        | Author  | Comment                                         *
 *  ---------------------------------------------------------------------   *
 *  24/07/2006  | Bahamut | Creation                                        *
 *  ---------------------------------------------------------------------   *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

#include "TextureManager.h"

/****************************************************************************
 *                                                                          *
 *    Function     : CTextureManager                                        * 
 *                                                                          *
 *    Description  : Class construtor                                       *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
CTextureManager::CTextureManager(void)
{
	m_TextureTable.clear();
	initialize();

	// Note: Texture Manager initialized.
}


/****************************************************************************
 *                                                                          *
 *    Function     : ~CTextureManager                                       * 
 *                                                                          *
 *    Description  : Class Destrutor                                        *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
CTextureManager::~CTextureManager(void)
{
	release();
	// Note: Texture Manager released.
}


/****************************************************************************
 *                                                                          *
 *    Function     : initialize                                             * 
 *                                                                          *
 *    Description  : Initializes the texture manager                        *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CTextureManager::initialize(void)
{
	GLuint id;
    release();

    glGenTextures( 1, &id );
    glBindTexture( GL_TEXTURE_2D, id );

    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );

    m_TextureTable["default"] = id;

    // Note: Default texture created here.
    int i, j, c;
    GLubyte *data = new GLubyte[ 64 * 64 * 4 ];

    for( i = 0; i < 64; i++ )
        for( j = 0; j < 64; j++ )
        {
            c = ( !(i & 8) ^ !(j & 8)) * 255;
            data[ (i * 256) + (j * 4) + 0 ] = (GLubyte)c;
            data[ (i * 256) + (j * 4) + 1 ] = (GLubyte)c;
            data[ (i * 256) + (j * 4) + 2 ] = (GLubyte)c;
            data[ (i * 256) + (j * 4) + 3 ] = (GLubyte)255;
        }

    glTexImage2D( GL_TEXTURE_2D, 0, 4, 64, 64, 0, GL_RGBA, GL_UNSIGNED_BYTE, data );
    delete [] data;
}


/****************************************************************************
 *                                                                          *
 *    Function     : load                                                   * 
 *                                                                          *
 *    Description  : Loads an image file                                    *
 *                                                                          *
 *    Interface    : I fileName = physical file to load                     *
 *                                                                          *
 *    Return Value : Return the texture identifiant if succeed else default *
 *                                                                          *
 ****************************************************************************/
GLuint CTextureManager::load(std::string fileName)
{
    return load(fileName, fileName);
}


/****************************************************************************
 *                                                                          *
 *    Function     : load                                                   * 
 *                                                                          *
 *    Description  : Loads an image file and affects an alias               *
 *                                                                          *
 *    Interface    : I name = alias name                                    *
 *                   I fileName = physical file to load                     *
 *                                                                          *
 *    Return Value : Texture index or default texture index if failed       *
 *                                                                          *
 ****************************************************************************/
GLuint CTextureManager::load(std::string name, std::string fileName)
{
    GLuint id = 0;

    id = index(name);

    if( id != index("default") )
    {
    }
    else
    {
		ILuint ImgId =0;
		ilGenImages(1, &ImgId);
		ilBindImage(ImgId);

		if (!ilLoadImage((ILstring)(fileName.c_str())))
		{
			// Error: Texture 'fileName' can't be loaded.
			return id;
		}

		ilEnable(IL_ORIGIN_SET);
		ilEnable(IL_FORMAT_SET);
		ilOriginFunc(IL_ORIGIN_UPPER_LEFT);
		ilutRenderer(ILUT_OPENGL);

		id = ilutGLBindTexImage();
		m_TextureTable[ name ] = id;

		ilDeleteImages(1, &ImgId);

		// Note: New texture created.
    }
    return id;
}


/****************************************************************************
 *                                                                          *
 *    Function     : index                                                  * 
 *                                                                          *
 *    Description  : Returns the index number                               *
 *                                                                          *
 *    Interface    : I name = texture alias                                 *
 *                                                                          *
 *    Return Value : Texture index or default texture index if failed       *
 *                                                                          *
 ****************************************************************************/
GLuint CTextureManager::index(std::string name)
{
    if( (m_Iter = m_TextureTable.find( name )) != m_TextureTable.end() )
        return (*m_Iter).second;
    else
        return m_TextureTable[ "default" ];
}


/****************************************************************************
 *                                                                          *
 *    Function     : name                                                   * 
 *                                                                          *
 *    Description  : Returns the texture alias                              *
 *                                                                          *
 *    Interface    : I id = texture identifiant                             *
 *                                                                          *
 *    Return Value : Texture alias or empty string if failed                *
 *                                                                          *
 ****************************************************************************/
const char* CTextureManager::name(GLuint id)
{
	if( glIsTexture( id ) )
    {
        for( m_Iter = m_TextureTable.begin(); m_Iter != m_TextureTable.end(); ++m_Iter )
        {
            if( (*m_Iter).second == id )
            {
                return (char *)((*m_Iter).first.c_str());
            }
        }
    }
    return (char *)"";
}


/****************************************************************************
 *                                                                          *
 *    Function     : release                                                * 
 *                                                                          *
 *    Description  : Frees alls textures (memories too)                     *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CTextureManager::release(void)
{
    for( m_Iter = m_TextureTable.begin(); m_Iter != m_TextureTable.end(); )
    {
        glDeleteTextures( 1, &(*m_Iter).second );
        m_TextureTable.erase( m_Iter++ );
    }
}

/****************************************************************************
 *                                                                          *
 *    Function     : del                                                    * 
 *                                                                          *
 *    Description  : Deletes one texture                                    *
 *                                                                          *
 *    Interface    : I name = alias of the texture to free                  *
 *                                                                          *
 *    Return Value : Returns true if succeed, else returns false            *
 *                                                                          *
 ****************************************************************************/
bool CTextureManager::del(std::string name)
{
    if( (m_Iter = m_TextureTable.find( name )) != m_TextureTable.end() )
    {
        glDeleteTextures( 1, &(*m_Iter).second );
        m_TextureTable.erase( m_Iter );
		// Note: Texture deleted.
		return true;
    }
	// Error: Can't delete texture.
	return false;
}


/****************************************************************************
 *                                                                          *
 *    Function     : del                                                    * 
 *                                                                          *
 *    Description  : Deletes one texture                                    *
 *                                                                          *
 *    Interface    : I id = identifier of the texture to free               *
 *                                                                          *
 *    Return Value : Returns true if succeed, else returns false            *
 *                                                                          *
 ****************************************************************************/
bool CTextureManager::del(GLuint id)
{
    if( glIsTexture( id ) )
    {
        for( m_Iter = m_TextureTable.begin(); m_Iter != m_TextureTable.end(); ++m_Iter )
        {
            if( (*m_Iter).second == id )
            {
                glDeleteTextures( 1, &(*m_Iter).second );
                m_TextureTable.erase( m_Iter );
				// Note: Texture with identifier id deleted.
                return true;
            }
        }
    }
	// Error: Can't delete texture.
	return false;
}


/****************************************************************************
 *                                                                          *
 *    Function     : bind                                                   * 
 *                                                                          *
 *    Description  : Binds one texture                                      *
 *                                                                          *
 *    Interface    : I name = alias of the texture to bind                  *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CTextureManager::bind(std::string name)
{
	glBindTexture(GL_TEXTURE_2D, index(name));
}


/****************************************************************************
 *                                                                          *
 *    Function     : bind                                                   * 
 *                                                                          *
 *    Description  : Binds one texture                                      *
 *                                                                          *
 *    Interface    : I id = identifiant of the texture to bind              *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CTextureManager::bind(GLuint id)
{
	glBindTexture(GL_TEXTURE_2D, id);
}


/****************************************************************************
 *                                                                          *
 *    Function     : readFile                                               * 
 *                                                                          *
 *    Description  : Loads a textures list in one file(XML Formalism)       *
 *                                                                          *
 *    Interface    : I fileName = physical file                             *
 *                                                                          *
 *    Return Value : Returns true if succeed, else returns false            *
 *                                                                          *
 ****************************************************************************/
bool CTextureManager::readFile(std::string fileName)
{
	TiXmlDocument doc( fileName );
	if ( !doc.LoadFile() )
	{
		// Error: File 'fileName' can't be parsed.
		return false;
	}

	TiXmlHandle hdl(&doc);
	TiXmlElement *XMLelement = hdl.FirstChildElement().FirstChildElement().Element();
 
	while (XMLelement)
	{
		std::string name = XMLelement->Attribute("name");
		std::string file = XMLelement->Attribute("file");

		load(name, file);

		XMLelement = XMLelement->NextSiblingElement();
	}
	return true;
}


/****************************************************************************
 *                                                                          *
 *    Function     : capture                                                * 
 *                                                                          *
 *    Description  : Takes a screen capture  (format TGA)                   *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CTextureManager::capture()
{
	ilutGLScreenie();
}



/****************************************************************************
 *                                                                          *
 *  NOTES :                                                                 *
 *                                                                          *
 ****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *    
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *         
 *                                                                          *    
 ****************************************************************************/ 
 