/****************************************************************************
 *                                                                          *
 *                  Concours 2006 - GAMES CREATORS NETWORK                  *
 *                                                                          *
 ****************************************************************************

                     PRODUCT          : GRAVITY GAME
 
                     FILE             : PhysicEngine.cpp

                     VERSION          : 1
 
 ****************************************************************************
 *                                                                          *
 *  DESCRIPTION :                                                           *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          * 
 *                                                                          *  
 ****************************************************************************
 *                                                                          *
 * HISTORY :                                                                * 
 * ---------                                                                *
 *                                                                          *
 *  Date        | Author  | Comment                                         *
 *  ---------------------------------------------------------------------   *
 *  25/07/2006  | Bahamut | Creation                                        *
 *  ---------------------------------------------------------------------   *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

#include "PhysicEngine.h"
#include "../global.h" 

void Callback(void *data,dGeomID o1, dGeomID o2)
{
	CPhysicEngine::m_pEngine->NearCallback(data,o1,o2);
}

CPhysicEngine* CPhysicEngine::m_pEngine = 0;

/****************************************************************************
 *                                                                          *
 *    Function     : CPhysicEngine                                          * 
 *                                                                          *
 *    Description  : Class construtor                                       *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
CPhysicEngine::CPhysicEngine(void)
{
	m_iSpace = 0;
	m_iWorld = 0;
	m_bInited = false;
	m_iCurrentIndexGenerator = 0;
	CPhysicEngine::m_pEngine = this;
}

/****************************************************************************
 *                                                                          *
 *    Function     : ~CPhysicEngine                                         * 
 *                                                                          *
 *    Description  : Class destructor                                       *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
CPhysicEngine::~CPhysicEngine(void)
{
	if (m_bInited)
		Close();
}

/****************************************************************************
 *                                                                          *
 *    Function     : Initialize                                             * 
 *                                                                          *
 *    Description  : Initialises the engine                                 *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :  Returns true if succeed                               *
 *                                                                          *
 ****************************************************************************/
bool CPhysicEngine::Initialize(void)
{
	bool result = false;
	if (!m_bInited)
	{
		m_bInited = true;
		result = true;
		m_iWorld = dWorldCreate();
		SetGravity(0.0f, -9.81f, 0.0f);
		SetPrecisionERP(0.4f);
		SetCorrectionCFM(0.000001f);
		m_iSpace = dHashSpaceCreate(0);
		m_iJointGroup = dJointGroupCreate(500);
	}

	return result;
}


/****************************************************************************
 *                                                                          *
 *    Function     : generateObjectIndex                                    * 
 *                                                                          *
 *    Description  : Generates the next object index;	                    *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
int CPhysicEngine::GenerateObjectIndex()
{
	return ++m_iCurrentIndexGenerator;
}

/****************************************************************************
 *                                                                          *
 *    Function     : SetGravity                                             * 
 *                                                                          *
 *    Description  : Sets gravity in the world  (-9.81f for ur earth        *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CPhysicEngine::SetGravity(float x, float y, float z)
{
	if (m_bInited)
		dWorldSetGravity(m_iWorld, x, y, z);
	else
		fprintf(stderr, "Error : Can't set gravity because Physic Engine isn't initialized.\n");
}

/****************************************************************************
 *                                                                          *
 *    Function     : SetPrecisionERP                                        * 
 *                                                                          *
 *    Description  : Sets Precision ERP                                     *
 *                                                                          *
 *    Interface    : I value : value between 0.1 and 0.8                    *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CPhysicEngine::SetPrecisionERP(float value)
{
	if (m_bInited)
		dWorldSetERP(m_iWorld, value);
	else
		fprintf(stderr, "Error : Can't set Correction ERP because Physic Engine isn't initialized.\n");
}


/****************************************************************************
 *                                                                          *
 *    Function     : SetCorrectionCFM                                       *
 *                                                                          *
 *    Description  : Sets Correction Joints Correction                      *
 *                                                                          *
 *    Interface    : I value : value between 10e-9 and 1                    *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CPhysicEngine::SetCorrectionCFM(float value)
{
	if (m_bInited)
		dWorldSetCFM(m_iWorld, value);
	else
		fprintf(stderr, "Error : Can't set Correction CFM because Physic Engine isn't initialized.\n");
}

/****************************************************************************
 *                                                                          *
 *    Function     : Close                                                  * 
 *                                                                          *
 *    Description  : Terminates the engine                                  *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CPhysicEngine::Close(void)
{
	m_bInited = false;
	dJointGroupEmpty(m_iJointGroup);
	dSpaceDestroy(m_iSpace);
	dWorldDestroy(m_iWorld);
	dCloseODE();
}

/****************************************************************************
 *                                                                          *
 *    Function     : Update                                                 * 
 *                                                                          *
 *    Description  : Updates simulation of one step                         *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CPhysicEngine::Update(void)
{
	dWorldStep(m_iWorld,0.04f);
	dJointGroupEmpty(m_iJointGroup);

	dSpaceCollide (m_iSpace, 0, Callback);
	
	/*const int N = 10;
	dContact contact[N];
	int n = dCollide (geometries[1], geometries[9], N, &contact[0].geom, sizeof(dContact));
	if (n>0)
	{
		printf("Collision de la bille sur le champignon !!\n");
		dGeomDisable(geometries[9]);
		dBodyAddForce(bodies[1],rand()%400-200,rand()%200+200,-rand()%100-100);
	}*/


	std::map<int, PObject *>::const_iterator itr;
	for(itr = m_vObjects.begin(); itr != m_vObjects.end(); ++itr)
	{
		PObject *obj = (*itr).second;
		if ( !obj->isStatic )
		{
			const dReal* position = dBodyGetPosition(obj->bodyID);
			obj->x = position[0];
			obj->y = position[1];
			obj->z = position[2];
		}
	}
}

/****************************************************************************
 *                                                                          *
 *    Function     : NearCallbac                                            * 
 *                                                                          *
 *    Description  :                                                        *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CPhysicEngine::NearCallback(void *data,dGeomID o1, dGeomID o2)
{
	int i, n;

	const int N = m_vObjects.size();
	dContact contact[100];
	n = dCollide (o1, o2, N, &contact[0].geom, sizeof(dContact));
	if (n > 0)
	{
		for (i=0; i<n; i++)
		{
			contact[i].surface.mode = dContactBounce;
			contact[i].surface.mu = 5000; //5000;
			contact[i].surface.bounce = 0.8;//0.3
			contact[i].surface.bounce_vel = 0.2;//0.2
  			dJointID c = dJointCreateContact (m_iWorld,m_iJointGroup,&contact[i]);
			dJointAttach(c,dGeomGetBody(o1), dGeomGetBody(o2));
		}
	}
}


/****************************************************************************
 *                                                                          *
 *    Function     :                                                        * 
 *                                                                          *
 *    Description  :                                                        *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
int CPhysicEngine::CreateSphere(bool isStatic, float radius)
{
	PObject *obj = new PObject();
	obj->type = PObject::Sphere;
	obj->radius = radius;
	obj->identifier = GenerateObjectIndex();
	obj->isStatic = isStatic;
	if (isStatic)
	{
		obj->geomID = dCreateSphere(m_iSpace,obj->radius);
		dGeomSetPosition(obj->geomID, obj->x, obj->y, obj->z);
		dGeomSetBody (obj->geomID,0);
	}
	else
	{
		dMass mass;
		obj->bodyID = dBodyCreate(m_iWorld);
		dBodySetPosition(obj->bodyID, obj->x, obj->y, obj->z);
		dBodyEnable(obj->bodyID);
		dBodySetGravityMode(obj->bodyID,1);
		dMassSetSphere(&mass, obj->mass, obj->radius);
		dBodySetMass(obj->bodyID, &mass);
		dBodySetForce(obj->bodyID,obj->fx,obj->fy,obj->fz);
		obj->geomID = dCreateSphere(m_iSpace,obj->radius);
		dGeomSetBody (obj->geomID,obj->bodyID);
	}

	m_vObjects[obj->identifier] = obj;
	return obj->identifier;
}


/****************************************************************************
 *                                                                          *
 *    Function     :                                                        * 
 *                                                                          *
 *    Description  :                                                        *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
int CPhysicEngine::CreateBox(bool isStatic, float x, float y, float z)
{
	PObject *obj = new PObject();
	obj->type = PObject::Box;
	obj->l = x;
	obj->L = y;
	obj->h = z;
	obj->identifier = GenerateObjectIndex();
	obj->isStatic = isStatic;
	if (isStatic)
	{
		obj->geomID = dCreateBox(m_iSpace, obj->l,obj->L,obj->h);
		dGeomSetPosition(obj->geomID, obj->x, obj->y, obj->z);
		dGeomSetBody (obj->geomID,0);
	}
	else
	{
		dMass mass;
		obj->bodyID = dBodyCreate(m_iWorld);
		dBodySetPosition(obj->bodyID, obj->x, obj->y, obj->z);
		dBodyEnable(obj->bodyID);
		dBodySetGravityMode(obj->bodyID,1);
		dMassSetBox(&mass, obj->mass, obj->l,obj->L,obj->h);
		dBodySetMass(obj->bodyID, &mass);
		dBodySetForce(obj->bodyID,obj->fx,obj->fy,obj->fz);
		obj->geomID = dCreateBox(m_iSpace,obj->l,obj->L,obj->h);
		dGeomSetBody (obj->geomID,obj->bodyID);
	}

	m_vObjects[obj->identifier] = obj;
	return obj->identifier;
}


/****************************************************************************
 *                                                                          *
 *    Function     :                                                        * 
 *                                                                          *
 *    Description  :                                                        *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
int CPhysicEngine::CreatePlane(float a, float b, float c, float d)
{
	PObject *obj = new PObject();
	obj->type = PObject::Plane;
	obj->a = a;
	obj->b = b;
	obj->c = c;
	obj->d = d;
	obj->identifier = GenerateObjectIndex();
	obj->isStatic = true;

	obj->geomID = dCreatePlane(m_iSpace, obj->a,obj->b,obj->c, obj->d);
	dGeomSetPosition(obj->geomID, obj->x, obj->y, obj->z);
	dGeomSetBody (obj->geomID,0);

	m_vObjects[obj->identifier] = obj;
	return obj->identifier;
}


/****************************************************************************
 *                                                                          *
 *    Function     :                                                        * 
 *                                                                          *
 *    Description  :                                                        *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
int CPhysicEngine::CreateCylinder(bool isStatic, float radius, float length)
{
	PObject *obj = new PObject();
	obj->type = PObject::Cylinder;
	obj->radius = radius;
	obj->l = length;
	obj->identifier = GenerateObjectIndex();
	obj->isStatic = isStatic;

	if (isStatic)
	{
		obj->geomID = dCreateCCylinder(m_iSpace, obj->radius,obj->l);
		dGeomSetPosition(obj->geomID, obj->x, obj->y, obj->z);
		dGeomSetBody (obj->geomID,0);
	}
	else
	{
		dMass mass;
		obj->bodyID = dBodyCreate(m_iWorld);
		dBodySetPosition(obj->bodyID, obj->x, obj->y, obj->z);
		dBodyEnable(obj->bodyID);
		dBodySetGravityMode(obj->bodyID,1);
		dMassSetCylinder(&mass, obj->mass, obj->radius,1.0f,obj->l);
		dBodySetMass(obj->bodyID, &mass);
		dBodySetForce(obj->bodyID,obj->fx,obj->fy,obj->fz);
		obj->geomID = dCreateCCylinder(m_iSpace,obj->radius,obj->l);
		dGeomSetBody (obj->geomID,obj->bodyID);
	}

	m_vObjects[obj->identifier] = obj;
	return obj->identifier;
}


/****************************************************************************
 *                                                                          *
 *    Function     :                                                        * 
 *                                                                          *
 *    Description  :                                                        *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CPhysicEngine::SetPosition(int id, float x, float y, float z)
{
	PObject *obj = m_vObjects[id];
	obj->x = x;
	obj->y = y;
	obj->z = z;
	if (obj->isStatic)
		dGeomSetPosition(obj->geomID, obj->x, obj->y, obj->z);
	else
		dBodySetPosition(obj->bodyID, obj->x, obj->y, obj->z);
}


/****************************************************************************
 *                                                                          *
 *    Function     :                                                        * 
 *                                                                          *
 *    Description  :                                                        *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CPhysicEngine::SetMass(int id, float mass)
{
	PObject *obj = m_vObjects[id];

	if (obj->isStatic)
		return;

	dMass dmass;
	obj->mass = mass;

	switch(obj->type)
	{
	case PObject::Box:
		dMassSetBox(&dmass, obj->mass, obj->l,obj->L,obj->h);
		dBodySetMass(obj->bodyID, &dmass);
		break;
	case PObject::Sphere:
		dMassSetSphere(&dmass, obj->mass, obj->radius);
		dBodySetMass(obj->bodyID, &dmass);
		break;
	case PObject::Cylinder:
	case PObject::CappedCylinder:
		dMassSetCylinder(&dmass, obj->mass, obj->radius,1.0f,obj->l);
		dBodySetMass(obj->bodyID, &dmass);
		break;
	case PObject::Plane:
		break;
	}
}



/****************************************************************************
 *                                                                          *
 *    Function     :                                                        * 
 *                                                                          *
 *    Description  :                                                        *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CPhysicEngine::AddForce(int id, float x, float y, float z)
{
	PObject *obj = m_vObjects[id];

	if (obj->isStatic)
		return;

	obj->fx += x;
	obj->fy += y;
	obj->fz += z;

	dBodyAddForce(obj->bodyID,x,y,z);
}


/****************************************************************************
 *                                                                          *
 *    Function     :                                                        * 
 *                                                                          *
 *    Description  :                                                        *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CPhysicEngine::SetForce(int id, float x, float y, float z)
{
	PObject *obj = m_vObjects[id];

	if (obj->isStatic)
		return;

	obj->fx = x;
	obj->fy = y;
	obj->fz = z;

	dBodySetForce(obj->bodyID,obj->fx,obj->fy,obj->fz);
}


/****************************************************************************
 *                                                                          *
 *    Function     :                                                        * 
 *                                                                          *
 *    Description  :                                                        *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CPhysicEngine::AddTorque(int id, float x, float y, float z)
{
	PObject *obj = m_vObjects[id];

	if (obj->isStatic)
		return;

	obj->tx += x;
	obj->ty += y;
	obj->tz += z;

	dBodyAddTorque(obj->bodyID,x,y,z);
}



/****************************************************************************
 *                                                                          *
 *    Function     :                                                        * 
 *                                                                          *
 *    Description  :                                                        *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CPhysicEngine::SetTorque(int id, float x, float y, float z)
{
	PObject *obj = m_vObjects[id];

	if (obj->isStatic)
		return;

	obj->tx = x;
	obj->ty = y;
	obj->tz = z;

	dBodySetTorque(obj->bodyID,obj->fx,obj->fy,obj->fz);
}


/****************************************************************************
 *                                                                          *
 *    Function     :                                                        * 
 *                                                                          *
 *    Description  :                                                        *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CPhysicEngine::SetEnabledGravity(int id, bool enable)
{
	PObject *obj = m_vObjects[id];

	if (obj->isStatic)
		return;

	if (enable)
		dBodySetGravityMode(obj->bodyID,1);
	else
		dBodySetGravityMode(obj->bodyID,0);
}



/****************************************************************************
 *                                                                          *
 *    Function     :                                                        * 
 *                                                                          *
 *    Description  :                                                        *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CPhysicEngine::SetRotation(int id, float angle, float x, float y, float z)
{
	PObject *obj = m_vObjects[id];

	dReal angleRad = (dReal)angle * 0.0174532925;
	dMatrix3  matrix;
	dRFromAxisAndAngle(matrix, x, y, z, angleRad);

	if (obj->isStatic)
		dGeomSetRotation(obj->geomID, matrix);
	else
		dBodySetRotation(obj->bodyID, matrix);
}


/****************************************************************************
 *                                                                          *
 *    Function     :                                                        * 
 *                                                                          *
 *    Description  :                                                        *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
PObject* CPhysicEngine::Object(int id)
{
	return (PObject*) m_vObjects[id];
}


/****************************************************************************
 *                                                                          *
 *    Function     :                                                        * 
 *                                                                          *
 *    Description  :                                                        *
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CPhysicEngine::Clean()
{
	dJointGroupEmpty(m_iJointGroup); //destroy contact groups

	std::map<int, PObject *>::const_iterator itr;
	for(itr = m_vObjects.begin(); itr != m_vObjects.end(); ++itr)
	{
		if (!(*itr).second->isStatic)
			dBodyDestroy(((*itr).second)->bodyID);
		dGeomDestroy(((*itr).second)->geomID);
	}

	m_vObjects.clear();
}

/****************************************************************************
 *                                                                          *
 *    Function     :  Draw                                                  * 
 *                                                                          *
 *    Description  :  Draws primitives for debugging                        * 
 *                                                                          *
 *    Interface    :                                                        *
 *                                                                          *
 *    Return Value :                                                        *
 *                                                                          *
 ****************************************************************************/
void CPhysicEngine::Draw(void)
{
	std::map<int, PObject *>::const_iterator itr;
	for(itr = m_vObjects.begin(); itr != m_vObjects.end(); ++itr)
	{
		PObject *obj = (*itr).second;
		const dReal* position;
		const dReal* rotation;
		if (obj->isStatic)
		{
			position = dGeomGetPosition(obj->geomID);
			rotation = dGeomGetRotation(obj->geomID);
		}
		else
		{
			position = dBodyGetPosition(obj->bodyID);
			rotation = dBodyGetRotation(obj->bodyID);
		}

		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		glColor3f(1.0f,0.0f,0.0f);
		glDisable(GL_TEXTURE);
		switch(obj->type)
		{
		case PObject::Box:
			{
			float l = obj->l;
			float L = obj->L;
			float h = obj->h;
			float x = position[0]-l/2.0f;
			float y = position[1]-L/2.0f;
			float z = position[2]-h/2.0f;
			glBegin(GL_QUADS);
				glVertex3f(x,y+L,7.0f);
				glVertex3f(x,y,7.0f);
				glVertex3f(x+l,y,7.0f);
				glVertex3f(x+l,y+L,7.0f);
			glEnd();

			break;
			}
		case PObject::Sphere:
			break;
		case PObject::Cylinder:
			break;
		case PObject::CappedCylinder:
			break;
		case PObject::Plane:
			break;
		}

		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	}
	


}