/****************************************************************************
 *                                                                          *
 *                  Concours 2006 - GAMES CREATORS NETWORK                  *
 *                                                                          *
 ****************************************************************************

                     PRODUCT          : GRAVITY GAME
 
                     FILE             : Vector.h

                     VERSION          : 1
 
 ****************************************************************************
 *                                                                          *
 *  DESCRIPTION :                                                           *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          * 
 *                                                                          *  
 ****************************************************************************
 *                                                                          *
 * HISTORY :                                                                * 
 * ---------                                                                *
 *                                                                          *
 *  Date        | Author  | Comment                                         *
 *  ---------------------------------------------------------------------   *
 *  24/07/2006  | Bahamut | Creation                                        *
 *  ---------------------------------------------------------------------   *
 *                                                                          *
 *                                                                          *
 ****************************************************************************/

#ifndef _VECTOR_H__
#define _VECTOR_H__

#include <math.h>
#define VECT_EPSILON 1e-6f

// ==============================================
// CVector2D - Vecteur 2D
// ==============================================
template<class T> class CVector2D
{
public:
	inline CVector2D(): x(0), y(0) {}
	inline CVector2D(T X, T Y):x(X), y(Y){}
	inline CVector2D(const T *v) : x(v[0]), y(v[1]) { }
	inline CVector2D(const CVector2D &v) : x(v.x), y(v.y) { }

	inline int operator==(const CVector2D &v) { return (fabs(x - v.x) < VECT_EPSILON && fabs(y - v.y) < VECT_EPSILON ); }
	inline int operator!=(const CVector2D &v) { return !(*this == v); }
	
	inline const CVector2D operator*(T f) const { return CVector2D(x * f,y * f); }
	inline const CVector2D operator/(T f) const { return CVector2D(x / f,y / f); }
	inline const CVector2D operator+(const CVector2D &v) const { return CVector2D(x + v.x,y + v.y); }
	inline const CVector2D operator-() const { return CVector2D(-x,-y); }
	inline const CVector2D operator-(const CVector2D &v) const { return CVector2D(x - v.x,y - v.y); }
	
	inline CVector2D &operator*=(T f) { return *this = *this * f; }
	inline CVector2D &operator/=(T f) { return *this = *this / f; }
	inline CVector2D &operator+=(const CVector2D &v) { return *this = *this + v; }
	inline CVector2D &operator-=(const CVector2D &v) { return *this = *this - v; }
	
	inline T operator*(const CVector2D &v) const { return x * v.x + y * v.y; }
	
	inline operator T*() { return (T*)&x; }
	inline operator const T*() const { return (T*)&x; }
	
	inline T &operator[](int i) { return ((T*)&x)[i]; }
	inline const T operator[](int i) const { return ((T*)&x)[i]; }
	
	inline float length() const { return (float)sqrt(x * x + y * y); }
	inline float normalize() {
		float inv,length = (float)sqrt(x * x + y * y);
		if(length < VECT_EPSILON) return 0.0f;
		inv = 1 / length;
		x *= inv;
		y *= inv;
		return length;
	}

	inline CVector2D<T> saturate() {
		CVector2D ret = CVector2D(x,y);
		if(ret.x < (T)0) ret.x = (T)0;
		else if(ret.x > (T)1) ret.x = (T)1;
		if(ret.y < (T)0) ret.y = (T)0;
		else if(ret.y > (T)1) ret.y = (T)1;
		return ret;
	}

	union {
		struct {
			T x,y;
		};
		T v[2];
	};				
};

// ==============================================
// CVector3D - Vector 3D
// ==============================================
template<class T> class CVector3D
{
public:
	inline CVector3D(): x(0), y(0),z(0) {}
	inline CVector3D(T X, T Y, T Z):x(X), y(Y), z(Z) {}
	inline CVector3D(const T *v) : x(v[0]), y(v[1]), z(v[2]) { }
	inline CVector3D(const CVector3D &v) : x(v.x), y(v.y), z(v.z) { }

	inline int operator==(const CVector3D &v) { return (fabs(x - v.x) < VECT_EPSILON && fabs(y - v.y) < VECT_EPSILON && fabs(z - v.z) < VECT_EPSILON); }
	inline int operator!=(const CVector3D &v) { return !(*this == v); }
	
	inline const CVector3D operator*(T f) const { return CVector3D(x * f,y * f, z * f); }
	inline const CVector3D operator/(T f) const { return CVector3D(x / f,y / f, z / f); }
	inline const CVector3D operator+(const CVector3D &v) const { return CVector3D(x + v.x,y + v.y, z + v.z); }
	inline const CVector3D operator-() const { return CVector3D(-x,-y,-z); }
	inline const CVector3D operator-(const CVector3D &v) const { return CVector3D(x - v.x,y - v.y,z - v.z); }
	
	inline CVector3D &operator*=(T f) { return *this = *this * f; }
	inline CVector3D &operator/=(T f) { return *this = *this / f; }
	inline CVector3D &operator+=(const CVector3D &v) { return *this = *this + v; }
	inline CVector3D &operator-=(const CVector3D &v) { return *this = *this - v; }
	
	inline T operator*(const CVector3D &v) const { return x * v.x + y * v.y + z * v.z; }
	
	inline operator T*() { return (T*)&x; }
	inline operator const T*() const { return (T*)&x; }
	
	inline T &operator[](int i) { return ((T*)&x)[i]; }
	inline const T operator[](int i) const { return ((T*)&x)[i]; }
	
	inline float length() const { return (float)sqrt(x * x + y * y + z * z); }
	inline float normalize() {
		float inv,length = (float)sqrt(x * x + y * y + z * z);
		if(length < VECT_EPSILON) return 0.0f;
		inv = 1 / length;
		x *= inv;
		y *= inv;
		z *= inv;
		return length;
	}

	inline void cross(const CVector3D &v1,const CVector3D &v2) {
		x = v1.y * v2.z - v1.z * v2.y;
		y = v1.z * v2.x - v1.x * v2.z;
		z = v1.x * v2.y - v1.y * v2.x;
	}

	inline CVector3D<T> saturate() {
		CVector3D ret = CVector3D(x,y,z);
		if(ret.x < (T)0) ret.x = (T)0;
		else if(ret.x > (T)1) ret.x = (T)1;
		if(ret.y < (T)0) ret.y = (T)0;
		else if(ret.y > (T)1) ret.y = (T)1;
		if(ret.z < (T)0) ret.z = (T)0;
		else if(ret.z > (T)1) ret.z = (T)1;
		return ret;
	}

	union {
		struct {
			T x,y,z;
		};
		T v[3];
	};				
};


// ==============================================
// CVector4D - Vector 4D
// ==============================================
template<class T> class CVector4D
{
public:
	inline CVector4D(): x(0), y(0),z(0), w(1) {}
	inline CVector4D(T X, T Y, T Z, T W):x(X), y(Y), z(Z), w(W) {}
	inline CVector4D(const T *v) : x(v[0]), y(v[1]), z(v[2]), w(v[3]) { }
	inline CVector4D(const CVector3D &v) : x(v.x), y(v.y), z(v.z), w(1) { }
	inline CVector4D(const CVector3D &v,T W) : x(v.x), y(v.y), z(v.z), w(W) { }
	inline CVector4D(const CVector4D &v) : x(v.x), y(v.y), z(v.z), w(v.w) { }

	inline int operator==(const CVector4D &v) { return (fabs(x - v.x) < VECT_EPSILON && fabs(y - v.y) < VECT_EPSILON && fabs(z - v.z) < VECT_EPSILON && fabs(w - v.w) < VECT_EPSILON); }
	inline int operator!=(const CVector4D &v) { return !(*this == v); }
	
	inline const CVector4D operator*(T f) const { return CVector4D(x * f,y * f, z * f, w * f); }
	inline const CVector4D operator/(T f) const { return CVector4D(x / f,y / f, z / f, w / f); }
	inline const CVector4D operator+(const CVector4D &v) const { return CVector4D(x + v.x,y + v.y, z + v.z, w + v.w); }
	inline const CVector4D operator-() const { return CVector4D(-x,-y,-z, -w); }
	inline const CVector4D operator-(const CVector4D &v) const { return CVector4D(x - v.x,y - v.y,z - v.z, w -v.w); }
	
	inline CVector4D &operator*=(T f) { return *this = *this * f; }
	inline CVector4D &operator/=(T f) { return *this = *this / f; }
	inline CVector4D &operator+=(const CVector4D<T> &v) { return *this = *this + v; }
	inline CVector4D &operator-=(const CVector4D<T> &v) { return *this = *this - v; }
	
	inline T operator*(const CVector3D<T> &v) const { return x * v.x + y * v.y + z * v.z + w; }
	inline T operator*(const CVector4D<T> &v) const { return x * v.x + y * v.y + z * v.z + w * v.w; }
	
	inline operator T*() { return (T*)&x; }
	inline operator const T*() const { return (T*)&x; }
	
	inline T &operator[](int i) { return ((T*)&x)[i]; }
	inline const T operator[](int i) const { return ((T*)&x)[i]; }

	union {
		struct {
			T x,y,z,w;
		};
		T v[4];
	};				
};


#endif


/****************************************************************************
 *                                                                          *
 *  NOTES :                                                                 *
 *                                                                          *
 ****************************************************************************
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *    
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *
 *                                                                          *         
 *                                                                          *    
 ****************************************************************************/